import os

import joblib
import merlin
import numpy as np
import xgboost as xgb
from joblib import dump
from merlin.model import PyFuncModel
from sklearn import svm
from sklearn.datasets import load_iris

XGB_PATH = os.path.join("models/", "model_1.bst")
SKLEARN_PATH = os.path.join("models/", "model_2.joblib")


class IrisModel(PyFuncModel):
    def initialize(self, artifacts):
        self._model_1 = xgb.Booster(model_file=artifacts["xgb_model"])
        self._model_2 = joblib.load(artifacts["sklearn_model"])

    def infer(self, model_input):
        inputs = np.array(model_input["instances"])
        dmatrix = xgb.DMatrix(model_input["instances"])
        result_1 = self._model_1.predict(dmatrix)
        result_2 = self._model_2.predict_proba(inputs)
        return {"predictions": ((result_1 + result_2) / 2).tolist()}


def train_models():
    iris = load_iris()
    y = iris["target"]
    X = iris["data"]

    # train xgboost model
    dtrain = xgb.DMatrix(X, label=y)
    param = {
        "max_depth": 6,
        "eta": 0.1,
        "silent": 1,
        "nthread": 4,
        "num_class": 3,
        "objective": "multi:softprob",
    }
    xgb_model = xgb.train(params=param, dtrain=dtrain)
    xgb_model.save_model(XGB_PATH)

    # train sklearn model
    clf = svm.SVC(gamma="scale", probability=True)
    clf.fit(X, y)
    dump(clf, SKLEARN_PATH)


if __name__ == "__main__":
    train_models()

    # test pyfunc model locally
    iris_model = IrisModel()
    iris_model.initialize(
        artifacts={
            "xgb_model": XGB_PATH,
            "sklearn_model": SKLEARN_PATH,
        }
    )
    pred = iris_model.infer({"instances": [[2.8, 1.0, 6.8, 0.4], [3.1, 1.4, 4.5, 1.6]]})
    print(pred)

    # run pyfunc model locally
    merlin.run_pyfunc_model(
        model_instance=IrisModel(),
        conda_env="env.yaml",
        artifacts={
            "xgb_model": XGB_PATH,
            "sklearn_model": SKLEARN_PATH,
        },
    )



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from prometheus_client.exposition import choose_encoder
from tornado.web import RequestHandler


class MetricsHandler(RequestHandler):
    def initialize(self, metrics_registry) -> None:
        self.metrics_registry = metrics_registry

    def get(self):
        encoder, content_type = choose_encoder(self.request.headers.get('accept'))
        self.set_header("Content-Type", content_type)
        self.write(encoder(self.metrics_registry))



import logging

import merlin
from merlin.model import PyFuncModel


class EchoModel(PyFuncModel):
    def initialize(self, artifacts):
        pass

    def infer(self, request):
        logging.info("request: %s", request)
        return request


if __name__ == "__main__":
    # Run pyfunc model locally without uploading to Merlin server
    merlin.run_pyfunc_model(
        model_instance=EchoModel(),
        conda_env="env.yaml",
    )



import time
import requests


def wait_server_ready(url, timeout_second=10, tick_second=2):
    ellapsed_second = 0
    while (ellapsed_second < timeout_second):
        try:
            resp = requests.get(url)
            if resp.status_code == 200:
                return
        except Exception as e:
            print(f"{url} is not ready: {e}")

        time.sleep(tick_second)
        ellapsed_second += tick_second

    if ellapsed_second >= timeout_second:
        raise TimeoutError("server is not ready within specified timeout duration")



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import http
import json
import os
import subprocess
import time
from abc import abstractmethod

import mlflow.pyfunc
import pytest
import requests
import tornado.web
from prometheus_client import Counter, Gauge

from pyfuncserver.config import HTTP_PORT, ModelManifest, WORKERS
from pyfuncserver.model.model import PyFuncModel
from test.utils import wait_server_ready
from merlin.pyfunc import PYFUNC_MODEL_INPUT_KEY, PYFUNC_EXTRA_ARGS_KEY, PyFuncOutput


class NewModelImpl(mlflow.pyfunc.PythonModel):
    def load_context(self, context):
        self.initialize(context.artifacts)
        self._use_kwargs_infer = True

    def predict(self, context, model_input):
        extra_args = model_input.get(PYFUNC_EXTRA_ARGS_KEY, {})
        input = model_input.get(PYFUNC_MODEL_INPUT_KEY, {})
        if extra_args is not None:
            response = self._do_predict(input, **extra_args)
            return PyFuncOutput(http_response=response)

        response = self._do_predict(input)
        return PyFuncOutput(http_response=response)

    def _do_predict(self, model_input, **kwargs):
        if self._use_kwargs_infer:
            try:
                return self.infer(model_input, **kwargs)
            except TypeError as e:
                if "infer() got an unexpected keyword argument" in str(e):
                    print(
                        'Fallback to the old infer() method, got TypeError exception: {}'.format(e))
                    self._use_kwargs_infer = False
                else:
                    raise e

        return self.infer(model_input)

    @abstractmethod
    def initialize(self, artifacts: dict):
        """
        Implementation of PyFuncModel can specify initialization step which
        will be called one time during model initialization.
        :param artifacts: dictionary of artifacts passed to log_model method
        """
        pass

    @abstractmethod
    def infer(self, request: dict, **kwargs) -> dict:
        """
        Do inference
        This method MUST be implemented by concrete implementation of
        PyFuncModel.
        This method accept 'request' which is the body content of incoming
        request.
        Implementation should return inference a json object of response.
        :param request: Dictionary containing incoming request body content
        :return: Dictionary containing response body
        """
        pass


class ModelImpl(mlflow.pyfunc.PythonModel):
    def load_context(self, context):
        self.initialize(context.artifacts)
        self._use_kwargs_infer = True

    def predict(self, model_input, **kwargs):
        if self._use_kwargs_infer:
            try:
                response = self.infer(model_input, **kwargs)
                return PyFuncOutput(http_response=response)
            except TypeError as e:
                if "infer() got an unexpected keyword argument" in str(e):
                    print(
                        'Fallback to the old infer() method, got TypeError exception: {}'.format(e))
                    self._use_kwargs_infer = False
                else:
                    raise e

        response = self.infer(model_input)
        return PyFuncOutput(http_response=response)

    @abstractmethod
    def initialize(self, artifacts: dict):
        """
        Implementation of PyFuncModel can specify initialization step which
        will be called one time during model initialization.
        :param artifacts: dictionary of artifacts passed to log_model method
        """
        pass

    @abstractmethod
    def infer(self, request: dict, **kwargs) -> dict:
        """
        Do inference
        This method MUST be implemented by concrete implementation of
        PyFuncModel.
        This method accept 'request' which is the body content of incoming
        request.
        Implementation should return inference a json object of response.
        :param request: Dictionary containing incoming request body content
        :return: Dictionary containing response body
        """
        pass


class EchoModel(ModelImpl):
    def initialize(self, artifacts):
        self._req_count = Counter("my_req_count", "Number of incoming request")
        self._temp = Gauge("my_gauge", "Number of incoming request")

    def infer(self, model_input):
        self._req_count.inc()
        self._temp.set(10)
        return model_input


class NewEchoModel(NewModelImpl):
    def initialize(self, artifacts):
        self._req_count = Counter("new_my_req_count", "Number of incoming request")
        self._temp = Gauge("new_my_gauge", "Number of incoming request")

    def infer(self, model_input):
        self._req_count.inc()
        self._temp.set(10)
        return model_input


class NonEmptyEchoModel(NewModelImpl):
    def initialize(self, artifacts):
        self._req_count = Counter("non_empty_my_req_count", "Number of incoming request")
        self._temp = Gauge("non_empty_my_gauge", "Number of incoming request")

    def infer(self, model_input):
        if model_input == {}:
            raise TypeError("Request could not be empty")
        self._req_count.inc()
        self._temp.set(10)
        return model_input


class HeadersModel(ModelImpl):
    def initialize(self, artifacts):
        pass

    def infer(self, model_input, **kwargs):
        return kwargs.get('headers', {})


class NewHeadersModel(NewModelImpl):
    def initialize(self, artifacts):
        pass

    def infer(self, model_input, **kwargs):
        return kwargs.get('headers', {})


class HttpErrorModel(ModelImpl):
    def initialize(self, artifacts):
        pass

    def infer(self, model_input):
        raise tornado.web.HTTPError(
            status_code=model_input["status_code"],
            reason=model_input["reason"]
        )


class NewHttpErrorModel(NewModelImpl):
    def initialize(self, artifacts):
        pass

    def infer(self, model_input):
        raise tornado.web.HTTPError(
            status_code=model_input["status_code"],
            reason=model_input["reason"]
        )

@pytest.mark.parametrize(
    "model",
    [EchoModel(), NewEchoModel(), NonEmptyEchoModel()]
)
def test_model(model):
    mlflow.pyfunc.log_model("model", python_model=model)
    model_path = os.path.join(mlflow.get_artifact_uri(), "model")
    model_manifest = ModelManifest(model_name="echo-model",
                                   model_version="1",
                                   model_full_name="echo-model-1",
                                   model_dir=model_path,
                                   project="sample")

    model = PyFuncModel(model_manifest)
    model.load()

    assert model.ready

    inputs = [[1, 2, 3], [4, 5, 6]]
    outputs = model.predict(inputs)
    assert outputs == PyFuncOutput(http_response=inputs)

@pytest.mark.parametrize(
    "model",
    [EchoModel(), NewEchoModel(), NonEmptyEchoModel()]
)
def test_model_int(model):
    mlflow.pyfunc.log_model("model", python_model=model)
    model_path = os.path.join(mlflow.get_artifact_uri(), "model")
    env = os.environ.copy()
    mlflow.end_run()

    try:
        env["PROMETHEUS_MULTIPROC_DIR"] = "prometheus"
        env[HTTP_PORT[0]] = "8081"
        env[WORKERS[0]] = "1"
        c = subprocess.Popen(["python", "-m", "pyfuncserver", "--model_dir", model_path], env=env)

        # wait till the server is up
        wait_server_ready("http://localhost:8081/")

        for request_file_json in os.listdir("benchmark"):
            if not request_file_json.endswith(".json"):
                continue

            with open(os.path.join("benchmark", request_file_json), "r") as f:
                req = json.load(f)

            resp = requests.post("http://localhost:8081/v1/models/model-1:predict",
                                 json=req)
            assert resp.status_code == 200
            assert req == resp.json()
            assert resp.headers["content-type"] == "application/json; charset=UTF-8"
    finally:
        c.kill()


@pytest.mark.parametrize(
    "model",
    [HeadersModel(), NewHeadersModel()]
)
def test_model_headers(model):
    mlflow.pyfunc.log_model("model", python_model=model)
    model_path = os.path.join(mlflow.get_artifact_uri(), "model")
    env = os.environ.copy()
    mlflow.end_run()

    try:
        env["PROMETHEUS_MULTIPROC_DIR"] = "prometheus"
        env[HTTP_PORT[0]] = "8081"
        env[WORKERS[0]] = "1"
        c = subprocess.Popen(["python", "-m", "pyfuncserver", "--model_dir", model_path], env=env)

        # wait till the server is up
        wait_server_ready("http://localhost:8081/")

        with open(os.path.join("benchmark", "small.json"), "r") as f:
            req = json.load(f)

        resp = requests.post("http://localhost:8081/v1/models/model-1:predict",
                             json=req, headers={'Foo': 'bar'})
        assert resp.status_code == 200
        assert resp.json()["Foo"] == "bar"
        assert resp.headers["content-type"] == "application/json; charset=UTF-8"
    finally:
        c.kill()


@pytest.mark.parametrize(
    "error_core,message,model",
    [
        (http.HTTPStatus.BAD_REQUEST, "invalid request", HttpErrorModel()),
        (http.HTTPStatus.NOT_FOUND, "not found", HttpErrorModel()),
        (http.HTTPStatus.INTERNAL_SERVER_ERROR, "server error", HttpErrorModel()),
(http.HTTPStatus.BAD_REQUEST, "invalid request", NewHttpErrorModel()),
        (http.HTTPStatus.NOT_FOUND, "not found", NewHttpErrorModel()),
        (http.HTTPStatus.INTERNAL_SERVER_ERROR, "server error", NewHttpErrorModel()),
    ]
)
def test_error_model_int(error_core, message, model):
    mlflow.pyfunc.log_model("model", python_model=model)
    model_path = os.path.join(mlflow.get_artifact_uri(), "model")
    env = os.environ.copy()
    mlflow.end_run()

    try:
        env["PROMETHEUS_MULTIPROC_DIR"] = "prometheus"
        env[HTTP_PORT[0]] = "8081"
        env[WORKERS[0]] = "1"
        c = subprocess.Popen(["python", "-m", "pyfuncserver", "--model_dir", model_path], env=env)

        # wait till the server is up
        wait_server_ready("http://localhost:8081/")
        
        req = {"status_code": error_core.value, "reason": message}
        resp = requests.post("http://localhost:8081/v1/models/model-1:predict",
                             json=req)
        assert resp.status_code == error_core.value
        assert resp.json() == {"status_code": error_core.value, "reason": message}
        assert resp.headers["content-type"] == "application/json; charset=UTF-8"
    finally:
        c.kill()





import pytest
from unittest.mock import patch, MagicMock
from datetime import datetime

from test.test_http import sample_model_input, sample_model_output
from pyfuncserver.config import ModelManifest
from pyfuncserver.utils.converter import build_prediction_log, _build_model_input, _build_model_output
from merlin.pyfunc import PyFuncOutput

from caraml.upi.v1 import prediction_log_pb2
from google.protobuf import timestamp_pb2


@patch("pyfuncserver.utils.converter.datetime_now")
def test_build_prediction_log(datetime_now):
    mock_datetime = datetime(year=2023, month=11, day=17, hour=1, minute=1, second=0, microsecond=2)
    datetime_now.return_value = mock_datetime
    pyfunc_output = PyFuncOutput(
        http_response = {"response": "ok"},
        model_input=sample_model_input,
        model_output=sample_model_output
    )
    manifest = ModelManifest(model_name="model_name", model_version="1", model_full_name="model_name_1", model_dir="/", project="project")
    got_prediction_log = build_prediction_log(pyfunc_output, manifest)
    exp_timestamp = timestamp_pb2.Timestamp()
    exp_timestamp.FromDatetime(mock_datetime)
    exp_prediction_log = prediction_log_pb2.PredictionLog(
            prediction_id=sample_model_input.session_id,
            target_name="",
            project_name="project",
            model_name="model_name",
            model_version="1",
            input= _build_model_input(pyfunc_output.model_input),
            output= _build_model_output(pyfunc_output.model_output),
            table_schema_version=1,
            request_timestamp=exp_timestamp
    )
    assert got_prediction_log == exp_prediction_log



from pyfuncserver.sampler.sampler import Sampler
from merlin.pyfunc import PyFuncOutput
from abc import ABC, abstractmethod
import asyncio

class Producer(ABC):

    @abstractmethod
    def produce(self, data: PyFuncOutput):
        pass

class Publisher:
    def __init__(self, producer: Producer, sampler: Sampler) -> None:
        self.producer = producer
        self.sampler = sampler

    async def publish(self, output: PyFuncOutput):
        if not self.sampler.should_sample():
            return

        self.producer.produce(output)


import logging
import os
import pathlib
import re
import shutil
import signal
import subprocess
import time

import grpc
import mlflow
import pandas as pd
import pytest
import requests
from caraml.upi.utils import df_to_table
from caraml.upi.v1 import upi_pb2, upi_pb2_grpc, variable_pb2, type_pb2, table_pb2
from grpc_health.v1 import health_pb2_grpc, health_pb2
from merlin.model import PyFuncModel, PyFuncV3Model
from merlin.pyfunc import ModelInput, ModelOutput, Values
from merlin.protocol import Protocol
from prometheus_client import Counter, Gauge

from pyfuncserver.config import GRPC_PORT, HTTP_PORT, MODEL_FULL_NAME, MODEL_NAME, MODEL_VERSION, PROTOCOL, WORKERS
from test.utils import wait_server_ready
from test.test_http import sample_model_input, sample_model_output


class EchoUPIModel(PyFuncModel):
    GAUGE_VALUE = 42

    def __init__(self, model_name, model_version):
        self._model_name = model_name
        self._model_version = model_version

    def initialize(self, artifacts: dict):
        self._req_count = Counter("request_count", "Number of incoming request")
        self._temp = Gauge("some_gauge", "Number of incoming request")

    def upiv1_infer(self, request: upi_pb2.PredictValuesRequest,
                    context: grpc.ServicerContext) -> upi_pb2.PredictValuesResponse:

        self._req_count.inc()
        self._temp.set(EchoUPIModel.GAUGE_VALUE)

        return upi_pb2.PredictValuesResponse(
            prediction_result_table=request.prediction_table,
            target_name=request.target_name,
            prediction_context=request.prediction_context,
            metadata=upi_pb2.ResponseMetadata(
                prediction_id=request.metadata.prediction_id,
                # TODO: allow user to get model name and version from PyFuncModel
                models=[upi_pb2.ModelMetadata(name=self._model_name, version=self._model_version)]
            )
        )

class SampleUPIPyFuncV3(PyFuncV3Model):

    def __init__(self, model_name, model_version):
        self._model_name = model_name
        self._model_version = model_version

    def upiv1_preprocess(self, request: upi_pb2.PredictValuesRequest,
                    context: grpc.ServicerContext) -> ModelInput:
        return sample_model_input
    
    def infer(self, model_input: ModelInput) -> ModelOutput:
        return sample_model_output
    
    def upiv1_postprocess(self, model_output: ModelOutput, request: upi_pb2.PredictValuesRequest) -> upi_pb2.PredictValuesResponse:
        return upi_pb2.PredictValuesResponse(
            target_name=request.target_name,
            prediction_result_table=table_pb2.Table(
                name="prediction_result_table",
                columns=[
                    table_pb2.Column(name="prediction_score", type=type_pb2.TYPE_DOUBLE),
                    table_pb2.Column(name="prediction_label", type=type_pb2.TYPE_STRING)
                ],
                rows=[
                    table_pb2.Row(row_id="0", values=[table_pb2.Value(double_value=0.95), table_pb2.Value(string_value="complete")]),
                    table_pb2.Row(row_id="1", values=[table_pb2.Value(double_value=0.43), table_pb2.Value(string_value="incomplete")]),
                    table_pb2.Row(row_id="2", values=[table_pb2.Value(double_value=0.59), table_pb2.Value(string_value="complete")]),
                ]
            ),
            prediction_context=request.prediction_context,
            metadata=upi_pb2.ResponseMetadata(
                models=[upi_pb2.ModelMetadata(name=self._model_name, version=self._model_version)],
                prediction_id=request.metadata.prediction_id
        )
    )

@pytest.mark.parametrize("workers", [(1), (4), (8)])
@pytest.mark.benchmark
def test_benchmark_multiprocess(workers, benchmark):
    model_name = "my-model"
    model_version = "1"
    grpc_port = 9001
    http_port = 8081
    target_name = "echo"
    metrics_path = "metrics_test"

    try:
        c = start_upi_server(EchoUPIModel(model_name, model_version), model_name, model_version, http_port, grpc_port, workers, metrics_path)

        channel = grpc.insecure_channel(f'localhost:{grpc_port}')
        
        health_check_stub = health_pb2_grpc.HealthStub(channel)
        response: health_pb2.HealthCheckResponse = health_check_stub.Check(health_pb2.HealthCheckRequest())
        assert response.status == health_pb2.HealthCheckResponse.SERVING

        stub = upi_pb2_grpc.UniversalPredictionServiceStub(channel)
        df = pd.DataFrame([[4, 1, "hi"]] * 3,
                          columns=['int_value', 'int_value_2', 'string_value'],
                          index=["0000", "1111", "2222"])
        prediction_id = "12345"

        prediction_context = [
            variable_pb2.Variable(name="int_context", type=type_pb2.TYPE_INTEGER, integer_value=1),
            variable_pb2.Variable(name="double_context", type=type_pb2.TYPE_DOUBLE, double_value=1.1),
            variable_pb2.Variable(name="string_context", type=type_pb2.TYPE_STRING, string_value="hello")
        ]

        def send_request():
            return stub.PredictValues(
                request=upi_pb2.PredictValuesRequest(prediction_table=df_to_table(df, "predict"),
                                                     target_name=target_name,
                                                     prediction_context=prediction_context,
                                                     metadata=upi_pb2.RequestMetadata(
                                                         prediction_id=prediction_id, )
                                                     )
            )

        response = benchmark(send_request)

        assert response.metadata.prediction_id == prediction_id
        assert response.metadata.models[0].name == model_name
        assert response.metadata.models[0].version == model_version
        assert list(response.prediction_context) == prediction_context
        assert response.target_name == target_name
        assert df_to_table(df, "predict") == response.prediction_result_table

    finally:
        os.killpg(os.getpgid(c.pid), signal.SIGTERM)
        shutil.rmtree(metrics_path)
        # Wait until the previous server have been terminated completely
        time.sleep(5)

@pytest.mark.parametrize("workers", [(1), (4)])
def test_upi(workers):
    model_name = "my-model"
    model_version = "1"
    grpc_port = 9001
    http_port = 8081
    target_name = "echo"
    metrics_path = "metrics_test"

    try:
        
        c = start_upi_server(EchoUPIModel(model_name, model_version), model_name, model_version, http_port, grpc_port, workers, metrics_path)

        channel = grpc.insecure_channel(f'localhost:{grpc_port}')

        health_check_stub = health_pb2_grpc.HealthStub(channel)
        response: health_pb2.HealthCheckResponse = health_check_stub.Check(health_pb2.HealthCheckRequest())
        assert response.status == health_pb2.HealthCheckResponse.SERVING

        stub = upi_pb2_grpc.UniversalPredictionServiceStub(channel)
        df = pd.DataFrame([[4, 1, "hi"]] * 3,
                          columns=['int_value', 'int_value_2', 'string_value'],
                          index=["0000", "1111", "2222"])
        prediction_id = "12345"

        prediction_context = [
            variable_pb2.Variable(name="int_context", type=type_pb2.TYPE_INTEGER, integer_value=1),
            variable_pb2.Variable(name="double_context", type=type_pb2.TYPE_DOUBLE, double_value=1.1),
            variable_pb2.Variable(name="string_context", type=type_pb2.TYPE_STRING, string_value="hello")
        ]
        response: upi_pb2.PredictValuesResponse = stub.PredictValues(
            request=upi_pb2.PredictValuesRequest(prediction_table=df_to_table(df, "predict"),
                                                 target_name=target_name,
                                                 prediction_context=prediction_context,
                                                 metadata=upi_pb2.RequestMetadata(
                                                     prediction_id=prediction_id, )
                                                 )
        )

        assert response.metadata.prediction_id == prediction_id
        assert response.metadata.models[0].name == model_name
        assert response.metadata.models[0].version == model_version
        assert list(response.prediction_context) == prediction_context
        assert response.target_name == target_name
        assert df_to_table(df, "predict") == response.prediction_result_table

        # test metrics
        resp = requests.get(f"http://localhost:{http_port}/metrics")
        assert resp.status_code == 200

        # Check request_count counter
        matches = re.findall(r"request_count_total\s(\d\.\d)", resp.text)
        assert len(matches) == 1
        assert 1 == int(float(matches[0]))

        # Check some_gauge gauge value
        matches = re.findall(r"some_gauge\{pid=\"\d+\"\}\s(\d+.\d+)", resp.text)
        assert len(matches) > 0
        for match in matches:
            gauge_value = int(float(match))
            assert gauge_value == EchoUPIModel.GAUGE_VALUE or gauge_value == 0

    finally:
        os.killpg(os.getpgid(c.pid), signal.SIGTERM)
        shutil.rmtree(metrics_path)
        # Wait until the previous server have been terminated completely
        time.sleep(5)


def test_pyfuncv3_upi():
    model_name = "my-model"
    model_version = "1"
    grpc_port = 9001
    http_port = 8081
    target_name = "echo"
    metrics_path = "metrics_test"

    try:
        c = start_upi_server(SampleUPIPyFuncV3(model_name, model_version),model_name, model_version, http_port, grpc_port, 1, metrics_path)

        channel = grpc.insecure_channel(f'localhost:{grpc_port}')

        health_check_stub = health_pb2_grpc.HealthStub(channel)
        response: health_pb2.HealthCheckResponse = health_check_stub.Check(health_pb2.HealthCheckRequest())
        assert response.status == health_pb2.HealthCheckResponse.SERVING

        stub = upi_pb2_grpc.UniversalPredictionServiceStub(channel)
        df = pd.DataFrame([[4, 1, "hi"]] * 3,
                          columns=['int_value', 'int_value_2', 'string_value'],
                          index=["0000", "1111", "2222"])
        prediction_id = "12345"

        prediction_context = [
            variable_pb2.Variable(name="int_context", type=type_pb2.TYPE_INTEGER, integer_value=1),
            variable_pb2.Variable(name="double_context", type=type_pb2.TYPE_DOUBLE, double_value=1.1),
            variable_pb2.Variable(name="string_context", type=type_pb2.TYPE_STRING, string_value="hello")
        ]
        response: upi_pb2.PredictValuesResponse = stub.PredictValues(
            request=upi_pb2.PredictValuesRequest(prediction_table=df_to_table(df, "predict"),
                                                 target_name=target_name,
                                                 prediction_context=prediction_context,
                                                 metadata=upi_pb2.RequestMetadata(
                                                     prediction_id=prediction_id, )
                                                 )
        )

        assert response.metadata.prediction_id == prediction_id
        assert response.metadata.models[0].name == model_name
        assert response.metadata.models[0].version == model_version
        assert list(response.prediction_context) == prediction_context
        assert response.target_name == target_name

        # test metrics
        resp = requests.get(f"http://localhost:{http_port}/metrics")
        assert resp.status_code == 200


    finally:
        os.killpg(os.getpgid(c.pid), signal.SIGTERM)
        shutil.rmtree(metrics_path)
        # Wait until the previous server have been terminated completely
        time.sleep(5)


def start_upi_server(python_model, model_name="my-model", model_version="1", http_port=8080, grpc_port=8081, workers=1, metrics_path="prometheus"):
    model_full_name = f"{model_name}-{model_version}"

    mlflow.pyfunc.log_model("model", python_model=python_model)
    model_path = os.path.join(mlflow.get_artifact_uri(), "model")
    env = os.environ.copy()
    mlflow.end_run()
    pathlib.Path(metrics_path).mkdir(exist_ok=True)

    env[PROTOCOL[0]] = Protocol.UPI_V1.value
    env[HTTP_PORT[0]] = str(http_port)
    env[GRPC_PORT[0]] = str(grpc_port)
    env[MODEL_NAME[0]] = model_name
    env[MODEL_VERSION[0]] = model_version
    env[MODEL_FULL_NAME[0]] = model_full_name
    env[WORKERS[0]] = str(workers)
    env["PROMETHEUS_MULTIPROC_DIR"] = metrics_path
    pid = subprocess.Popen(["python", "-m", "pyfuncserver", "--model_dir", model_path], env=env, start_new_session=True)

    # wait till the server is up
    wait_server_ready(f"http://localhost:{http_port}/")
    return pid


import logging
import multiprocessing
from concurrent import futures

import grpc
from grpc import aio
import asyncio
from typing import Optional
from caraml.upi.v1 import upi_pb2, upi_pb2_grpc
from grpc_reflection.v1alpha import reflection
from grpc_health.v1.health import HealthServicer
from grpc_health.v1 import health_pb2_grpc

from pyfuncserver.config import Config
from pyfuncserver.model.model import PyFuncModel
from pyfuncserver.publisher.publisher import Publisher
from pyfuncserver.publisher.kafka import KafkaProducer
from pyfuncserver.sampler.sampler import RatioSampling
from merlin.pyfunc import PyFuncOutput


class PredictionService(upi_pb2_grpc.UniversalPredictionServiceServicer):
    def __init__(self, model: PyFuncModel):
        if not model.ready:
            model.load()
        self._model = model
        self._publisher: Optional[Publisher] = None

    def set_publisher(self, publisher: Publisher):
        if self._publisher is None:
            self._publisher = publisher

    def PredictValues(self, request, context):
        output = self._model.upiv1_predict(request=request, context=context)
        upi_response = output
        output_is_pyfunc_output = isinstance(upi_response, PyFuncOutput) 
        if output_is_pyfunc_output:
            upi_response = output.upi_response

        if self._publisher is not None and output_is_pyfunc_output and output.contains_prediction_log():
        # need to also check whether the output contains prediction log since the pyfunc server doesn't know which model that is used
            asyncio.create_task(self._publisher.publish(output))

        return upi_response


class UPIServer:
    def __init__(self, model: PyFuncModel, config: Config):
        self._predict_service = PredictionService(model=model)
        self._config = config
        self._health_service = HealthServicer()

    def start(self):
        logging.info(f"Starting {self._config.workers} workers")

        if self._config.workers > 1:
            # multiprocessing based on https://github.com/grpc/grpc/tree/master/examples/python/multiprocessing
            workers = []
            for _ in range(self._config.workers - 1):
                worker = multiprocessing.Process(target=self._run_server)
                worker.start()
                workers.append(worker)
        
        if self._config.publisher is not None:
            kafka_producer = KafkaProducer(self._config.publisher, self._config.model_manifest)
            sampler = RatioSampling(self._config.publisher.sampling_ratio)
            publisher = Publisher(kafka_producer, sampler)
            self._predict_service.set_publisher(publisher)

        asyncio.get_event_loop().run_until_complete(self._run_server())

    async def _run_server(self):
        """
            Start a server in a subprocess.

        """
        options = self._config.grpc_options
        options.append(('grpc.so_reuseport', 1))

        server = aio.server(futures.ThreadPoolExecutor(max_workers=self._config.grpc_concurrency),
                             options=options)
        upi_pb2_grpc.add_UniversalPredictionServiceServicer_to_server(self._predict_service, server)
        health_pb2_grpc.add_HealthServicer_to_server(self._health_service, server)

        # Enable reflection server for debugging
        SERVICE_NAMES = (
            upi_pb2.DESCRIPTOR.services_by_name['UniversalPredictionService'].full_name,
            reflection.SERVICE_NAME,
        )
        reflection.enable_server_reflection(SERVICE_NAMES, server)

        logging.info(
            f"Starting grpc service at port {self._config.grpc_port} with options {self._config.grpc_options}")
        server.add_insecure_port(f"[::]:{self._config.grpc_port}")
        await server.start()
        await server.wait_for_termination()



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

import prometheus_client
from merlin.protocol import Protocol
from prometheus_client import CollectorRegistry, multiprocess

from pyfuncserver.config import Config
from pyfuncserver.metrics.pusher import labels, start_metrics_pusher
from pyfuncserver.model.model import PyFuncModel
from pyfuncserver.protocol.rest.server import HTTPServer
from pyfuncserver.protocol.upi.server import UPIServer

class PyFuncServer:
    def __init__(self, config: Config):
        self._config = config

    def start(self, model: PyFuncModel):
        # initialize prometheus
        # register to MultiProcessCollector as PyFuncServer will run in multiple process
        registry = CollectorRegistry()
        multiprocess.MultiProcessCollector(registry)

        # start only one server depending on the chosen protocol
        # the intent is to save memory consumption
        if self._config.protocol == Protocol.HTTP_JSON:
            http_server = HTTPServer(model=model, config=self._config, metrics_registry=registry)
            http_server.start()
        elif self._config.protocol == Protocol.UPI_V1:
            # Due to https://github.com/knative/serving/issues/8471, we have to resort to pushing metrics to
            # prometheus push gateway.
            if (self._config.push_gateway.enabled):
                target_info = labels(self._config)
                start_metrics_pusher(self._config.push_gateway.url,
                                     registry,
                                     target_info,
                                     self._config.push_gateway.push_interval_sec)
            else:
                # start prometheus metrics server and listen at http port
                logging.info(f"starting metrics server at {self._config.http_port}")
                prometheus_client.start_http_server(self._config.http_port, registry=registry)

            # start grpc/upi server and listen at grpc port
            upi_server = UPIServer(model=model, config=self._config)
            upi_server.start()
        else:
            raise ValueError(f"unknown protocol {self._config.protocol}")



import pytest
from unittest.mock import MagicMock
from pyfuncserver.publisher.publisher import Producer, Publisher
from pyfuncserver.sampler.sampler import Sampler
from test.test_http import sample_model_input, sample_model_output
from merlin.pyfunc import PyFuncOutput

class AlwaysSampling(Sampler):
    def should_sample(self, request_id: int = None) -> bool :
        return True
    
class NeverSampling(Sampler):
    def should_sample(self, request_id: int = None) -> bool :
        return False


@pytest.mark.asyncio
@pytest.mark.parametrize("sampler, is_publishing", [(AlwaysSampling(), True), (NeverSampling(), False)])
async def test_publisher(sampler, is_publishing):
    producer = MagicMock(spec=Producer)
    publisher = Publisher(producer, sampler)
    pyfunc_output = PyFuncOutput(
        http_response = {"response": "ok"},
        model_input=sample_model_input,
        model_output=sample_model_output
    )
    await publisher.publish(pyfunc_output)
    if is_publishing:
        producer.produce.assert_called_once()
    else:
        producer.produce.assert_not_called()



import asyncio
import logging
import signal

import tornado
from prometheus_client import CollectorRegistry
from pyfuncserver.config import Config
from pyfuncserver.metrics.handler import MetricsHandler
from pyfuncserver.model.model import PyFuncModel
from pyfuncserver.protocol.rest.handler import (HealthHandler, LivenessHandler,
                                                PredictHandler)
from pyfuncserver.publisher.kafka import KafkaProducer
from pyfuncserver.publisher.publisher import Publisher
from pyfuncserver.sampler.sampler import RatioSampling


async def sig_handler(server):
    logging.info("Stop accepting new connection")
    server.stop()
    logging.info("Closing existing connection")
    await server.close_all_connections()
    logging.info("Stopping ioloop")
    tornado.ioloop.IOLoop.current().stop()


class Application(tornado.web.Application):
    def __init__(
        self,
        config: Config,
        metrics_registry: CollectorRegistry,
        registered_models: dict,
    ):
        self.publisher = None
        self.model_manifest = config.model_manifest
        handlers = [
            # Server Liveness API returns 200 if server is alive.
            (r"/", LivenessHandler),
            # Model Health API returns 200 if model is ready to serve.
            (
                r"/v1/models/([a-zA-Z0-9_-]+)",
                HealthHandler,
                dict(models=registered_models),
            ),
            (
                r"/v1/models/([a-zA-Z0-9_-]+):predict",
                PredictHandler,
                dict(models=registered_models),
            ),
            (r"/metrics", MetricsHandler, dict(metrics_registry=metrics_registry)),
        ]
        super().__init__(handlers)  # type: ignore  # noqa


class HTTPServer:
    def __init__(
        self, model: PyFuncModel, config: Config, metrics_registry: CollectorRegistry
    ):
        self.config = config
        self.workers = config.workers
        self.model_manifest = config.model_manifest
        self.http_port = config.http_port
        self.metrics_registry = metrics_registry
        self.registered_models: dict = {}
        self.register_model(model)

    def create_application(self):
        return Application(self.config, self.metrics_registry, self.registered_models)

    def start(self):
        application = self.create_application()
        self._http_server = tornado.httpserver.HTTPServer(application)
        logging.info("Listening on port %s", self.http_port)
        self._http_server.bind(self.http_port)
        logging.info("Will fork %d workers", self.workers)
        self._http_server.start(self.workers)

        # kafka producer must be initialize after fork the process
        if self.config.publisher is not None:
            kafka_producer = KafkaProducer(
                self.config.publisher, self.config.model_manifest
            )
            sampler = RatioSampling(self.config.publisher.sampling_ratio)
            application.publisher = Publisher(kafka_producer, sampler)

        for signame in ("SIGINT", "SIGTERM"):
            asyncio.get_event_loop().add_signal_handler(
                getattr(signal, signame),
                lambda: asyncio.create_task(sig_handler(self._http_server)),
            )

        tornado.ioloop.IOLoop.current().start()

    def register_model(self, model: PyFuncModel):
        self.registered_models[model.full_name] = model
        logging.info(
            "Registering model: name: %s, version: %s, fullname: %s, predict endpoint: /v1/models/%s:predict",
            model.name,
            model.version,
            model.full_name,
            model.full_name,
        )



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import asyncio
import logging
import traceback

import uvloop
from pyfuncserver.config import Config
from pyfuncserver.model.model import PyFuncModel
from pyfuncserver.server import PyFuncServer
from pyfuncserver.utils.contants import ERR_DRY_RUN

parser = argparse.ArgumentParser()
parser.add_argument(
    "--model_dir", required=True, help="A URI pointer to the model binary"
)
parser.add_argument(
    "--dry_run",
    default=False,
    action="store_true",
    required=False,
    help="Dry run pyfunc server by loading the specified model "
    "in --model_dir without starting webserver",
)
args, _ = parser.parse_known_args()

logging.getLogger("tornado.access").disabled = True


def main():
    config = Config(args.model_dir)
    logging.basicConfig(level=config.log_level)

    # use uvloop as the event loop
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    # load model
    model = PyFuncModel(config.model_manifest)

    try:
        logging.info("Initializing model...")
        model.load()
        logging.info("Model initialized successfully.")
    except Exception as e:
        logging.error(f"Unable to initalize model")
        logging.error(traceback.format_exc())
        logging.error(ERR_DRY_RUN)
        exit(1)

    if args.dry_run:
        logging.info("dry run success")
        exit(0)

    logging.info("Starting server...")
    PyFuncServer(config).start(model)


if __name__ == "__main__":
    main()



import json
import logging
import os
from typing import Optional

from merlin.protocol import Protocol
from dataclasses import dataclass
from merlin.model_schema import ModelSchema

# Following environment variables are expected to be populated by Merlin
HTTP_PORT = ("CARAML_HTTP_PORT", 8080)
MODEL_NAME = ("CARAML_MODEL_NAME", "model")
MODEL_VERSION = ("CARAML_MODEL_VERSION", "1")
MODEL_FULL_NAME = ("CARAML_MODEL_FULL_NAME", "model-1")
PROJECT = ("CARAML_PROJECT", "project")
PROTOCOL = ("CARAML_PROTOCOL", "HTTP_JSON")
MODEL_SCHEMA = "CARAML_MODEL_SCHEMA"

WORKERS = ("WORKERS", 1)
GRPC_PORT = ("CARAML_GRPC_PORT", 9000)
LOG_LEVEL = ("LOG_LEVEL", "INFO")
GRPC_OPTIONS = ("GRPC_OPTIONS", "{}")
GRPC_CONCURRENCY = ("GRPC_CONCURRENCY", 10)

PUSHGATEWAY_ENABLED = ("PUSHGATEWAY_ENABLED", "false")
PUSHGATEWAY_URL = ("PUSHGATEWAY_URL", "localhost:9091")
PUSHGATEWAY_PUSH_INTERVAL_SEC = ("PUSHGATEWAY_PUSH_INTERVAL_SEC", 30)

PUBLISHER_KAFKA_TOPIC = ("PUBLISHER_KAFKA_TOPIC", "")
PUBLISHER_KAFKA_BROKERS = ("PUBLISHER_KAFKA_BROKERS", "")
PUBLISHER_KAKFA_LINGER_MS = ("PUBLISHER_KAFKA_LINGER_MS", 1000)
PUBLISHER_KAFKA_ACKS = ("PUBLISHER_KAFKA_ACKS", 0)
PUBLISHER_KAFKA_NUM_PARTITIONS = ("PUBLISHER_KAFKA_NUM_PARTITIONS", 24)
PUBLISHER_KAFKA_REPLICATION_FACTOR = ("PUBLISHER_KAFKA_REPLICATION_FACTOR", 3)
PUBLISHER_KAFKA_CONFIG = ("PUBLISHER_KAFKA_CONFIG", "{}")
PUBLISHER_SAMPLING_RATIO = ("PUBLISHER_SAMPLING_RATIO", 0.01)
PUBLISHER_ENABLED = ("PUBLISHER_ENABLED", "false")


@dataclass
class ModelManifest:
    """
    Model Manifest
    """
    model_name: str
    model_version: str
    model_full_name: str
    model_dir: str
    project: str
    model_schema: ModelSchema = None


class PushGateway:
    def __init__(self, enabled, url, push_interval_sec):
        self.url = url
        self.enabled = enabled
        self.push_interval_sec = push_interval_sec

@dataclass
class Kafka:
    """
    Kafka configuration
    """
    topic: str
    brokers: str
    linger_ms: int
    acks: int
    configuration: dict

@dataclass
class Publisher:
    """
    Publisher configuration
    """
    # sampling ratio of data that needs to be published
    sampling_ratio: float
    enabled: bool
    kafka: Kafka
    num_partitions: int
    replication_factor: int


class Config:
    """
    Server Configuration
    """

    def __init__(self, model_dir: str):
        self.protocol = Protocol(os.getenv(*PROTOCOL))
        self.http_port = int(os.getenv(*HTTP_PORT))
        self.grpc_port = int(os.getenv(*GRPC_PORT))

        # Model manifest
        model_name = os.getenv(*MODEL_NAME)
        model_version = os.getenv(*MODEL_VERSION)
        model_full_name = os.getenv(*MODEL_FULL_NAME)
        project = os.getenv(*PROJECT)
        model_schema = None
        model_schema_from_env_var = os.getenv(MODEL_SCHEMA) 
        if model_schema_from_env_var is not None:
            model_schema = ModelSchema.from_json(os.getenv(MODEL_SCHEMA))

        self.model_manifest = ModelManifest(model_name=model_name, 
                                            model_version=model_version, 
                                            model_full_name=model_full_name, 
                                            model_dir=model_dir, 
                                            project=project,
                                            model_schema=model_schema)

        self.workers = int(os.getenv(*WORKERS))
        self.log_level = self._log_level()

        self.grpc_options = self._grpc_options()
        self.grpc_concurrency = int(os.getenv(*GRPC_CONCURRENCY))

        push_enabled = str_to_bool(os.getenv(*PUSHGATEWAY_ENABLED))
        push_url = os.getenv(*PUSHGATEWAY_URL)
        push_interval = os.getenv(*PUSHGATEWAY_PUSH_INTERVAL_SEC)
        self.push_gateway = PushGateway(push_enabled,
                                        push_url,
                                        push_interval)
        
            
        # Publisher
        self.publisher = None
        publisher_enabled = str_to_bool(os.getenv(*PUBLISHER_ENABLED))
        if publisher_enabled:
            sampling_ratio = float(os.getenv(*PUBLISHER_SAMPLING_RATIO))
            kafka_topic = os.getenv(*PUBLISHER_KAFKA_TOPIC)
            kafka_brokers = os.getenv(*PUBLISHER_KAFKA_BROKERS)
            if kafka_topic == "":
                raise ValueError("kafka topic must be set")
            if kafka_brokers == "":
                raise ValueError("kafka brokers must be set")
            kafka_linger_ms = int(os.getenv(*PUBLISHER_KAKFA_LINGER_MS))
            kafka_acks = int(os.getenv(*PUBLISHER_KAFKA_ACKS))
            num_partitions = int(os.getenv(*PUBLISHER_KAFKA_NUM_PARTITIONS))
            replication_factor = int(os.getenv(*PUBLISHER_KAFKA_REPLICATION_FACTOR))
            kafka_cfgs = self._kafka_config()
            kafka = Kafka(
                kafka_topic, 
                kafka_brokers,
                kafka_linger_ms, 
                kafka_acks, 
                kafka_cfgs)
            self.publisher = Publisher(sampling_ratio, publisher_enabled, kafka, num_partitions, replication_factor)

    
    def _kafka_config(self):
        raw_cfg = os.getenv(*PUBLISHER_KAFKA_CONFIG)
        cfg = json.loads(raw_cfg)
        return cfg

    def _log_level(self):
        log_level = os.getenv(*LOG_LEVEL)
        numeric_level = getattr(logging, log_level.upper(), None)
        if not isinstance(numeric_level, int):
            logging.warning(f"invalid log level {log_level}")
            return logging.INFO
        return numeric_level

    def _grpc_options(self):
        raw_options = os.getenv(*GRPC_OPTIONS)
        options = json.loads(raw_options)
        grpc_options = []
        for k, v in options.items():
            grpc_options.append((k, v))
        return grpc_options

def str_to_bool(str: str)->bool:
    return str.lower() in ("true", "1")


import json
import pytest
import os
import pathlib
import re
import shutil
import signal
import subprocess
import random
import pandas as pd
import numpy as np

import mlflow
import requests
from merlin.model import PyFuncModel, PyFuncV3Model
from merlin.pyfunc import ModelInput, ModelOutput, Values, PyFuncOutput
from prometheus_client.metrics import Counter, Gauge

from pyfuncserver.config import HTTP_PORT, WORKERS, MODEL_NAME, MODEL_VERSION, MODEL_FULL_NAME
from test.utils import wait_server_ready


sample_model_input = ModelInput(
            prediction_ids=["prediction_1", "prediction_2", "prediction_3"],
            features=Values(
                columns=["featureA", "featureB", "featureC"],
                data=[[0.1, 0.2, 0.3], [0.2, 0.4, 0.6], [0.8, 0.5, 0.2]]
            ),
            entities=Values(
                columns=["order_id", "customer_id"],
                data=[["order1", "1111"], ["order1", "2222"], ["order1", "3333"]]
            ),
            session_id="randomSessionID"
        )

sample_model_input_np = ModelInput(
            prediction_ids=["prediction_1", "prediction_2", "prediction_3"],
            features=Values(
                columns=["featureA", "featureB", "featureC"],
                data=np.array([[0.1, 0.2, 0.3], [0.2, 0.4, 0.6], [0.8, 0.5, 0.2]])
            ),
            entities=Values(
                columns=["order_id", "customer_id"],
                data=np.array([["order1", "1111"], ["order1", "2222"], ["order1", "3333"]])
            ),
            session_id="randomSessionID"
        )

sample_model_input_df = ModelInput(
            prediction_ids=["prediction_1", "prediction_2", "prediction_3"],
            features = pd.DataFrame([[0.1, 0.2, 0.3], [0.2, 0.4, 0.6], [0.8, 0.5, 0.2]], columns=["featureA", "featureB", "featureC"]),
            entities = pd.DataFrame([["order1", "1111"], ["order1", "2222"], ["order1", "3333"]], columns=["order_id", "customer_id"]),
            session_id="randomSessionID"
        )
sample_model_output = ModelOutput(
            prediction_ids=["prediction_1", "prediction_2", "prediction3"],
            predictions=Values(
                columns=["prediction_score", "prediction_label"],
                data=[[0.9, "fraud"], [0.2, "not fraud"], [0.6, "fraud"]]
            )
        )
sample_model_output_np = ModelOutput(
            prediction_ids=["prediction_1", "prediction_2", "prediction3"],
            predictions=Values(
                columns=["prediction_score", "prediction_label"],
                data=np.array([[0.9, "fraud"], [0.2, "not fraud"], [0.6, "fraud"]])
            )
        )
sample_model_output_df = ModelOutput(
            prediction_ids=["prediction_1", "prediction_2", "prediction3"],
            predictions = pd.DataFrame([[0.9, "fraud"], [0.2, "not fraud"], [0.6, "fraud"]], columns=["prediction_score", "prediction_label"])
        )



class EchoModel(PyFuncModel):
    GAUGE_VALUE = 5

    def initialize(self, artifacts):
        self._req_count = Counter("request_count", "Number of incoming request")
        self._temp = Gauge("some_gauge", "Number of incoming request")

    def infer(self, model_input, **kwargs):
        if model_input == {}:
            raise TypeError("Request could not be empty")
        self._req_count.inc()
        self._temp.set(EchoModel.GAUGE_VALUE)
        return model_input
    
class SamplePyFuncV3Model(PyFuncV3Model):

    def preprocess(self, request: dict, **kwargs) -> ModelInput:
        return sample_model_input
    def infer(self, model_input: ModelInput) -> ModelOutput:
        return sample_model_output
    def postprocess(self, model_output: ModelOutput, request: dict) -> dict:
        return model_output.predictions_dict()
    
class SamplePyFuncV3NumpyModel(PyFuncV3Model):
    def preprocess(self, request: dict, **kwargs) -> ModelInput:
        return sample_model_input_np
    def infer(self, model_input: ModelInput) -> ModelOutput:
        return sample_model_output_np
    def postprocess(self, model_output: ModelOutput, request: dict) -> dict:
        return model_output.predictions_dict()
    
class SamplePyFuncV3DFModel(PyFuncV3Model):
    def preprocess(self, request: dict, **kwargs) -> ModelInput:
        return sample_model_input_df
    def infer(self, model_input: ModelInput) -> ModelOutput:
        return sample_model_output_df
    def postprocess(self, model_output: ModelOutput, request: dict) -> dict:
        return model_output.predictions_dict()


class SamplePyFuncWithModelObservability(PyFuncModel):
    def infer(self, request, **kwargs) -> ModelOutput:
        model_input = sample_model_input
        model_output = sample_model_output
        http_response = model_output.predictions_dict()
        return PyFuncOutput(http_response=http_response, model_input=model_input, model_output=model_output)
    

def test_http_protocol():
    mlflow.pyfunc.log_model("model", python_model=EchoModel())
    model_path = os.path.join(mlflow.get_artifact_uri(), "model")
    env = os.environ.copy()
    mlflow.end_run()

    model_name = "my-model"
    model_version = "2"
    model_full_name = f"{model_name}-{model_version}"
    port = "8081"

    try:
        env[HTTP_PORT[0]] = port
        env[MODEL_NAME[0]] = model_name
        env[MODEL_VERSION[0]] = model_version
        env[MODEL_FULL_NAME[0]] = model_full_name
        env[WORKERS[0]] = "1"
        env["PROMETHEUS_MULTIPROC_DIR"] = "prometheus"
        c = subprocess.Popen(["python", "-m", "pyfuncserver", "--model_dir", model_path], env=env)

        # wait till the server is up
        wait_server_ready(f"http://localhost:{port}/")

        for request_file_json in os.listdir("benchmark"):
            if not request_file_json.endswith(".json"):
                continue

            with open(os.path.join("benchmark", request_file_json), "r") as f:
                req = json.load(f)

            # test predict endpoint
            resp = requests.post(f"http://localhost:{port}/v1/models/{model_full_name}:predict",
                                 json=req)
            assert resp.status_code == 200
            assert req == resp.json()
            assert resp.headers["content-type"] == "application/json; charset=UTF-8"

            # test readiness endpoint
            resp = requests.get(f"http://localhost:{port}/")
            assert resp.status_code == 200

            # test readiness endpoint
            resp = requests.get(f"http://localhost:{port}/v1/models/{model_full_name}")
            assert resp.status_code == 200
    finally:
        c.kill()

@pytest.mark.parametrize("pyfunc_v3_model,expected", [
    (SamplePyFuncV3Model(), sample_model_output.predictions_dict()),
    (SamplePyFuncV3DFModel(), sample_model_output_df.predictions_dict()),
    (SamplePyFuncV3NumpyModel(), sample_model_output_np.predictions_dict())
])
def test_http_pyfuncv3(pyfunc_v3_model, expected):
    mlflow.pyfunc.log_model("model", python_model=pyfunc_v3_model)
    model_path = os.path.join(mlflow.get_artifact_uri(), "model")
    env = os.environ.copy()
    mlflow.end_run()

    model_name = "my-model"
    model_version = "2"
    model_full_name = f"{model_name}-{model_version}"
    port = "8081"

    try:
        env[HTTP_PORT[0]] = port
        env[MODEL_NAME[0]] = model_name
        env[MODEL_VERSION[0]] = model_version
        env[MODEL_FULL_NAME[0]] = model_full_name
        env[WORKERS[0]] = "1"
        env["PROMETHEUS_MULTIPROC_DIR"] = "prometheus"
        c = subprocess.Popen(["python", "-m", "pyfuncserver", "--model_dir", model_path], env=env)

        # wait till the server is up
        wait_server_ready(f"http://localhost:{port}/")

        for request_file_json in os.listdir("benchmark"):
            if not request_file_json.endswith(".json"):
                continue

            with open(os.path.join("benchmark", request_file_json), "r") as f:
                req = json.load(f)

            # test predict endpoint
            resp = requests.post(f"http://localhost:{port}/v1/models/{model_full_name}:predict",
                                 json=req)
            assert resp.status_code == 200
            assert resp.json() == expected
            assert resp.headers["content-type"] == "application/json; charset=UTF-8"

            # test readiness endpoint
            resp = requests.get(f"http://localhost:{port}/")
            assert resp.status_code == 200

            # test readiness endpoint
            resp = requests.get(f"http://localhost:{port}/v1/models/{model_full_name}")
            assert resp.status_code == 200
    finally:
        c.kill()

def test_http_pyfunc_with_model_observability():
    mlflow.pyfunc.log_model("model", python_model=SamplePyFuncWithModelObservability())
    model_path = os.path.join(mlflow.get_artifact_uri(), "model")
    env = os.environ.copy()
    mlflow.end_run()

    model_name = "my-model"
    model_version = "2"
    model_full_name = f"{model_name}-{model_version}"
    port = "8081"

    try:
        env[HTTP_PORT[0]] = port
        env[MODEL_NAME[0]] = model_name
        env[MODEL_VERSION[0]] = model_version
        env[MODEL_FULL_NAME[0]] = model_full_name
        env[WORKERS[0]] = "1"
        env["PROMETHEUS_MULTIPROC_DIR"] = "prometheus"
        c = subprocess.Popen(["python", "-m", "pyfuncserver", "--model_dir", model_path], env=env)

        # wait till the server is up
        wait_server_ready(f"http://localhost:{port}/")

        for request_file_json in os.listdir("benchmark"):
            if not request_file_json.endswith(".json"):
                continue

            with open(os.path.join("benchmark", request_file_json), "r") as f:
                req = json.load(f)

            # test predict endpoint
            resp = requests.post(f"http://localhost:{port}/v1/models/{model_full_name}:predict",
                                 json=req)
            assert resp.status_code == 200
            assert resp.json() == sample_model_output.predictions_dict()
            assert resp.headers["content-type"] == "application/json; charset=UTF-8"

            # test readiness endpoint``
            resp = requests.get(f"http://localhost:{port}/")
            assert resp.status_code == 200

            # test readiness endpoint
            resp = requests.get(f"http://localhost:{port}/v1/models/{model_full_name}")
            assert resp.status_code == 200
    finally:
        c.kill()


def test_metrics():
    mlflow.pyfunc.log_model("model", python_model=EchoModel())
    model_path = os.path.join(mlflow.get_artifact_uri(), "model")
    env = os.environ.copy()
    mlflow.end_run()

    model_name = "my-model"
    model_version = "1"
    model_full_name = f"{model_name}-{model_version}"
    port = "8082"
    metrics_path = "metrics_test"

    try:
        pathlib.Path(metrics_path).mkdir(exist_ok=True)

        env[HTTP_PORT[0]] = port
        env[MODEL_NAME[0]] = model_name
        env[MODEL_VERSION[0]] = model_version
        env[MODEL_FULL_NAME[0]] = model_full_name
        env[WORKERS[0]] = "4"
        env["PROMETHEUS_MULTIPROC_DIR"] = metrics_path
        c = subprocess.Popen(["python", "-m", "pyfuncserver", "--model_dir", model_path], env=env,
                             start_new_session=True)
        # wait till the server is up
        wait_server_ready(f"http://localhost:{port}/")

        predict_count = 0
        for request_file_json in os.listdir("benchmark"):
            if not request_file_json.endswith(".json"):
                continue

            with open(os.path.join("benchmark", request_file_json), "r") as f:
                req = json.load(f)

            # test predict endpoint
            resp = requests.post(f"http://localhost:{port}/v1/models/{model_full_name}:predict",
                                 json=req)
            assert resp.status_code == 200
            assert req == resp.json()
            assert resp.headers["content-type"] == "application/json; charset=UTF-8"
            predict_count += 1

        resp = requests.get(f"http://localhost:{port}/metrics")
        assert resp.status_code == 200

        # Check request_count counter
        matches = re.findall(r"request_count_total\s(\d\.\d)", resp.text)
        assert len(matches) == 1
        assert predict_count == int(float(matches[0]))

        # Check some_gauge gauge value
        matches = re.findall(r"some_gauge\{pid=\"\d+\"\}\s(\d+.\d+)", resp.text)
        assert len(matches) > 0
        for match in matches:
            gauge_value = int(float(match))
            assert gauge_value == EchoModel.GAUGE_VALUE or gauge_value == 0

    finally:
        os.killpg(os.getpgid(c.pid), signal.SIGTERM)
        shutil.rmtree(metrics_path)



import os
import random
import socket
import time
from multiprocessing import Process

import merlin
import pytest
import requests
from examples.iris_http.iris_http import IrisModel

request_json = {"instances": [[2.8, 1.0, 6.8, 0.4], [3.1, 1.4, 4.5, 1.6]]}

if os.environ.get("CI_SERVER"):
    host = "172.17.0.1"
else:
    host = "localhost"


def _get_free_port():
    sock = None
    try:
        while True:
            port = random.randint(8000, 9000)
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            result = sock.connect_ex(("127.0.0.1", port))
            if result != 0:
                return port
    finally:
        if sock is not None:
            sock.close()


def _wait_server_ready(proc, url, timeout_second=600, tick_second=10):
    time.sleep(5)

    ellapsed_second = 0
    while ellapsed_second < timeout_second:
        if not proc.is_alive():
            if proc.exitcode is not None and proc.exitcode != 0:
                raise RuntimeError("server failed to start")

        try:
            resp = requests.get(url)
            if resp.status_code == 200:
                return
        except Exception as e:
            print(f"{url} is not ready: {e}")

        time.sleep(tick_second)
        ellapsed_second += tick_second

    if ellapsed_second >= timeout_second:
        raise TimeoutError("server is not ready within specified timeout duration")


def _get_local_endpoint(model_full_name, port):
    return f"http://{host}:{port}/v1/models/{model_full_name}:predict"


@pytest.mark.skip(reason="need to release merlin-sdk==0.41.0 first")
@pytest.mark.local_server_test
def test_examples_iris():
    XGB_PATH = os.path.join("examples/iris_http/models/", "model_1.bst")
    SKLEARN_PATH = os.path.join("examples/iris_http/models/", "model_2.joblib")

    port = _get_free_port()

    p = Process(
        target=merlin.run_pyfunc_model,
        kwargs={
            "model_instance": IrisModel(),
            "conda_env": "examples/iris_http/env.yaml",
            "code_dir": ["examples"],
            "artifacts": {
                "xgb_model": XGB_PATH,
                "sklearn_model": SKLEARN_PATH,
            },
            "debug": True,
            "port": port,
        },
    )
    p.start()

    _wait_server_ready(p, f"http://{host}:{port}", timeout_second=600)

    resp = requests.post(_get_local_endpoint("irismodel-dev", port), json=request_json)
    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])
    p.terminate()



from pyfuncserver.config import Publisher as PublisherConfig, ModelManifest
from pyfuncserver.utils.converter import build_prediction_log

from confluent_kafka import KafkaError, KafkaException, Producer
from confluent_kafka.admin import AdminClient, NewTopic
from merlin.pyfunc import PyFuncOutput


class KafkaProducer(Producer):
    def __init__(self, publisher_config: PublisherConfig, model_manifest: ModelManifest) -> None:
        conf = {
            "bootstrap.servers": publisher_config.kafka.brokers, 
            "acks": publisher_config.kafka.acks, 
            "linger.ms": publisher_config.kafka.linger_ms
        }
        conf.update(publisher_config.kafka.configuration)
        self.producer = Producer(**conf)
        self.topic = publisher_config.kafka.topic
        self.model_manifest = model_manifest
        admin_client = AdminClient(conf)
        try:
            admin_client.create_topics([
                NewTopic(
                    self.topic,
                    num_partitions=publisher_config.num_partitions,
                    replication_factor=publisher_config.replication_factor
                )
            ])[self.topic].result()
        except KafkaException as e:
            if e.args[0].code() != KafkaError.TOPIC_ALREADY_EXISTS:
                raise e

    def produce(self, data: PyFuncOutput):
        prediction_log = build_prediction_log(pyfunc_output=data, model_manifest=self.model_manifest)
        serialized_data = prediction_log.SerializeToString() 
        self.producer.produce(topic=self.topic, value=serialized_data)
        self.producer.poll(0)



import pytest
import math

from pyfuncserver.sampler.sampler import RatioSampling

@pytest.mark.parametrize("num_requests, ratio", [(3000, 0.2), (10000, 0.3), (5000, 0.5)])
def test_ratio_sampling(num_requests, ratio):
    ratio_based_sampler = RatioSampling(ratio=ratio)
    num_sampled = 0
    for _ in range(num_requests):
        if ratio_based_sampler.should_sample():
            num_sampled = num_sampled + 1
    actual_ratio = num_sampled / num_requests
    print(f"actual ratio {actual_ratio} expected ratio {ratio}")
    assert math.isclose(a=ratio, b=actual_ratio, abs_tol=0.02)



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from enum import Enum

import grpc
from caraml.upi.v1 import upi_pb2

from merlin.protocol import Protocol
from merlin.pyfunc import PYFUNC_EXTRA_ARGS_KEY, PYFUNC_GRPC_CONTEXT, PYFUNC_MODEL_INPUT_KEY, PYFUNC_PROTOCOL_KEY, PyFuncOutput
from mlflow import pyfunc

from pyfuncserver.config import ModelManifest

NUM_OF_LATEST_PREDICT_FUNC_ARGS = 3


class PyFuncModelVersion(Enum):
    """
    PyFunc model version based on the merlin-sdk and mlflow version used in PyFunc model
    * LATEST -> using latest merlin-sdk or user doesn't specify merlin-sdk as dependency in PyFunc model
    * OLD_PYFUNC_LATEST_MLFLOW -> using older merlin-sdk (< 0.16) which doesn't have `predict(context, model_input)` signature,
        using latest version of mlflow (> 1.8).
    * OLD_PYFUNC_OLD_MLFLOW -> using older merlin-sdk (< 0.16) which doesn't have `predict(context, model_input)` signature,
        using older version of mlflow (<= 1.8)

    older version of merlin-sdk, merlin-sdk < 0.16
    older version of mlflow,  mlflow <= 1.8
    """
    LATEST = 'latest'
    OLD_PYFUNC_LATEST_MLFLOW = 'old_pyfunc_latest_mlflow'
    OLD_PYFUNC_OLD_MLFLOW = 'old_pyfunc_old_mlflow'


def _is_latest_pyfunc_model_version(func):
    full_args = inspect.getfullargspec(func)
    return len(full_args.args) == NUM_OF_LATEST_PREDICT_FUNC_ARGS


class PyFuncModel:  # pylint:disable=c-extension-no-member

    def __init__(self, model_manifest: ModelManifest):
        self.name = model_manifest.model_name
        self.version = model_manifest.model_version
        self.full_name = model_manifest.model_full_name
        self.artifact_dir = model_manifest.model_dir
        self.ready = False
        self.pyfunc_type = PyFuncModelVersion.LATEST

    def load(self):
        self._model = pyfunc.load_model(self.artifact_dir)
        # pylint:disable=c-extension-no-member
        # pylint:disable=attribute-defined-outside-init
        self.ready = True
        self.pyfunc_type = self._get_pyfunc_model_version()

    def _get_pyfunc_model_version(self):
        if hasattr(self._model, 'python_model'):
            if not _is_latest_pyfunc_model_version(self._model.python_model.predict):
                return PyFuncModelVersion.OLD_PYFUNC_OLD_MLFLOW
        elif hasattr(self._model, '_model_impl'):
            if not _is_latest_pyfunc_model_version(self._model._model_impl.python_model.predict):
                return PyFuncModelVersion.OLD_PYFUNC_LATEST_MLFLOW

        return PyFuncModelVersion.LATEST

    def predict(self, inputs: dict, **kwargs) -> PyFuncOutput:
        if self.pyfunc_type == PyFuncModelVersion.OLD_PYFUNC_LATEST_MLFLOW:
            # for case user specified old merlin-sdk as dependency and using mlflow without version specified
            return self._model._model_impl.python_model.predict(inputs, **kwargs)
        elif self.pyfunc_type == PyFuncModelVersion.OLD_PYFUNC_OLD_MLFLOW:
            # for case user specified old merlin-sdk as dependency and using old version of mlflow
            # that has `python_model` attribute on the model
            return self._model.python_model.predict(inputs, **kwargs)
        else:
            # for case user doesn't specify merlin-sdk as dependency
            model_inputs = {PYFUNC_MODEL_INPUT_KEY: inputs}
            if kwargs is not None:
                model_inputs[PYFUNC_EXTRA_ARGS_KEY] = kwargs

            return self._model.predict(model_inputs)

    def upiv1_predict(self, request: upi_pb2.PredictValuesRequest,
                      context: grpc.ServicerContext) -> PyFuncOutput:
        model_inputs = {
            PYFUNC_PROTOCOL_KEY: Protocol.UPI_V1,
            PYFUNC_MODEL_INPUT_KEY: request,
            PYFUNC_GRPC_CONTEXT: context,
        }
        return self._model.predict(model_inputs)



ERR_DRY_RUN = """
Some module might be missing from your pyfunc model.
If the module is part of your code, 
you can specify it as 'code' argument when calling log_pyfunc, e.g.:

merlin.log_pyfunc_model(model_instance=MyModel(), 
        conda_env="env.yaml",
        code=["path/to/your/package"])
        
If the module is a python package, then you should add it in the
conda env.yaml that you pass to log_pyfunc method, e.g.:

dependencies:
- python=3.8
- pip:
  - my-missing-package==1.0
  
OR if the package is not available in PyPI distribution

dependencies:
- python=3.8
- my-missing-package=1.0

See: https://docs.conda.io/projects/conda/en/latest/user-guide/tasks/manage-environments.html#creating-an-environment-from-an-environment-yml-file
"""



import pytest
from unittest.mock import MagicMock, patch, ANY
from pyfuncserver.publisher.kafka import KafkaProducer
from pyfuncserver.config import ModelManifest

from test.test_http import sample_model_input, sample_model_output
from merlin.pyfunc import PyFuncOutput
from test.test_http import sample_model_input, sample_model_output

@pytest.mark.parametrize("data", [
    PyFuncOutput(
        http_response = {"response": "ok"},
        model_input=sample_model_input,
        model_output=sample_model_output
    )
])
def test_kafka_producer(data):
    with patch.object(KafkaProducer, "__init__", lambda x, y, z: None):
        kafka_producer = KafkaProducer(None, None)
        producer = MagicMock()
        kafka_producer.producer = producer
        kafka_producer.topic = "sample_topic"
        kafka_producer.model_manifest = ModelManifest(model_name="model_name", model_version="1", model_full_name="model_name_1", model_dir="/", project="project")
        
        kafka_producer.produce(data)
        producer.produce.assert_called_once_with(topic="sample_topic", value=ANY)
        


import grpc
import pandas as pd
from caraml.upi.utils import df_to_table
from caraml.upi.v1 import upi_pb2, upi_pb2_grpc


def create_upi_request() -> upi_pb2.PredictValuesRequest:
    target_name = "echo"
    df = pd.DataFrame(
        [[4, 1, "hi"]] * 3,
        columns=["int_value", "int_value_2", "string_value"],
        index=["0000", "1111", "2222"],
    )
    prediction_id = "12345"

    return upi_pb2.PredictValuesRequest(
        target_name=target_name,
        prediction_table=df_to_table(df, "predict"),
        metadata=upi_pb2.RequestMetadata(prediction_id=prediction_id),
    )


if __name__ == "__main__":
    channel = grpc.insecure_channel(f"localhost:8080")
    stub = upi_pb2_grpc.UniversalPredictionServiceStub(channel)

    request = create_upi_request()
    response = stub.PredictValues(request=request)
    print(response)



import logging
import os

import grpc
import merlin
from caraml.upi.v1 import upi_pb2
from merlin.model import PyFuncModel
from merlin.protocol import Protocol
from prometheus_client import Counter, Gauge


class EchoUPIModel(PyFuncModel):
    GAUGE_VALUE = 42

    def __init__(self, model_name, model_version):
        self._model_name = model_name
        self._model_version = model_version

    def initialize(self, artifacts: dict):
        self._req_count = Counter("request_count", "Number of incoming request")
        self._temp = Gauge("some_gauge", "Number of incoming request")

    def upiv1_infer(
        self, request: upi_pb2.PredictValuesRequest, context: grpc.ServicerContext
    ) -> upi_pb2.PredictValuesResponse:
        logging.info(f"PID: {os.getpid()}")
        return upi_pb2.PredictValuesResponse(
            prediction_result_table=request.prediction_table,
            target_name=request.target_name,
            prediction_context=request.prediction_context,
            metadata=upi_pb2.ResponseMetadata(
                prediction_id=request.metadata.prediction_id,
                # TODO: allow user to get model name and version from PyFuncModel
                models=[
                    upi_pb2.ModelMetadata(
                        name=self._model_name, version=self._model_version
                    )
                ],
            ),
        )


if __name__ == "__main__":
    model_name = "echo-model"
    model_version = "1"

    merlin.run_pyfunc_model(
        model_instance=EchoUPIModel(model_name, model_version),
        conda_env="env.yaml",
        protocol=Protocol.UPI_V1,
    )



import random
from abc import ABC, abstractmethod
from typing import Optional

class Sampler(ABC):
    @abstractmethod
    def should_sample(self, request_id: Optional[int] = None) -> bool:
        pass

class RatioSampling(Sampler):
    """
    RatioSampling sampling strategy that mimic implementation of the TraceIDRatioBased in opentelemetry
    ref: https://github.com/open-telemetry/opentelemetry-python/blob/v1.21.0/opentelemetry-sdk/src/opentelemetry/sdk/trace/sampling.py#L253
    """
    REQUEST_ID_LIMIT = (1 << 64) - 1

    def __init__(self, ratio: float) -> None:
        if ratio < 0.0 or ratio > 1.0:
            raise ValueError("Valid ratio is on 0 - 1 range")
        self._ratio = ratio   
        self._bound = self.get_bound_for_ratio(ratio)

    @classmethod
    def get_bound_for_ratio(cls, ratio: float) -> int:
        return round(ratio * (cls.REQUEST_ID_LIMIT + 1))
    
    @property
    def ratio(self) -> float:
        return self._ratio
    
    def should_sample(self, request_id: Optional[int] = None) -> bool :
        if request_id is None:
            request_id = random.getrandbits(128)
        
        return request_id & self.REQUEST_ID_LIMIT < self._bound



import logging
import socket
import time
from threading import Thread

from prometheus_client import push_to_gateway

from pyfuncserver.config import Config


def start_metrics_pusher(push_gateway, registry, target_info, interval_sec):
    """
    Start periodic job to push metrics to prometheus push gatewat
    """
    logging.info(f"starting metrics pusher, url: {push_gateway} with interval {interval_sec} s")
    daemon = Thread(target=push_metrics, args=(push_gateway, registry, interval_sec, target_info),
                    daemon=True, name='metrics_push')
    daemon.start()


def push_metrics(gateway_url, registry, interval_sec, grouping_keys):
    """
    push metrics to prometheus push gateway every interval_sec
    Should be called in separate thread
    """
    while True:
        push_to_gateway(gateway_url, "merlin_pyfunc_upi", registry=registry, grouping_key=grouping_keys)
        time.sleep(interval_sec)


def labels(config: Config):
    """
    Labels to be added to all metrics
    """
    return {
        "merlin_model_name": config.model_manifest.model_name,
        "merlin_model_version": config.model_manifest.model_version,
        "host": socket.getfqdn()
    }



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import imp
import os

from setuptools import find_packages, setup

version = imp.load_source(
    "pyfuncserver.version", os.path.join("pyfuncserver", "version.py")
).VERSION

with open("requirements.txt") as f:
    REQUIRE = f.read().splitlines()

TESTS_REQUIRE = [
    "joblib>=0.13.0,<1.2.0",  # >=1.2.0 upon upgrade of kserve's version
    "mypy",
    "pytest-benchmark",
    "pytest-tornasync",
    "pytest",
    "requests",
    "scikit-learn>=1.1.2",
    "types-protobuf",
    "types-requests",
    "xgboost==1.6.2",
]

setup(
    name="merlin-pyfunc-server",
    version=version,
    author_email="merlin-dev@gojek.com",
    description="Model Server implementation for Merlin PyFunc model",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    python_requires=">=3.8,<3.11",
    packages=find_packages(exclude=["test"]),
    install_requires=REQUIRE,
    tests_require=TESTS_REQUIRE,
    extras_require={"test": TESTS_REQUIRE},
    entry_points="""
        [console_scripts]
        merlin-pyfunc-server=pyfuncserver.__main__:main
    """,
)



from datetime import datetime
from typing import Optional

from caraml.upi.v1 import prediction_log_pb2
from google.protobuf import timestamp_pb2
from google.protobuf.struct_pb2 import Struct
from merlin.pyfunc import ModelInput, ModelOutput, PyFuncOutput
from pyfuncserver.config import ModelManifest


def build_prediction_log(
    pyfunc_output: PyFuncOutput, model_manifest: ModelManifest
) -> prediction_log_pb2.PredictionLog:
    model_input = _build_model_input(pyfunc_output.model_input)
    model_output = _build_model_output(pyfunc_output.model_output)

    session_id = pyfunc_output.get_session_id()
    proto_timestamp = timestamp_pb2.Timestamp()
    proto_timestamp.FromDatetime(datetime_now())
    prediction_log = prediction_log_pb2.PredictionLog(
        prediction_id=session_id,
        target_name="",  # TO-DO update this after schema is introduced
        project_name=model_manifest.project,
        model_name=model_manifest.model_name,
        model_version=model_manifest.model_version,
        input=model_input,
        output=model_output,
        table_schema_version=1,
        request_timestamp=proto_timestamp,
    )
    return prediction_log


def datetime_now() -> datetime:
    return datetime.now()


def _new_struct(dictionary: Optional[dict]):
    struct = None
    if dictionary is not None:
        struct = Struct()
        struct.update(dictionary)

    return struct


def _build_model_input(input: ModelInput) -> prediction_log_pb2.ModelInput:
    features_table = _new_struct(input.features_dict()) if input is not None else None
    entities_table = _new_struct(input.entities_dict()) if input is not None else None
    return prediction_log_pb2.ModelInput(
        features_table=features_table, entities_table=entities_table
    )


def _build_model_output(output: ModelOutput) -> prediction_log_pb2.ModelOutput:
    prediction_results_table = (
        _new_struct(output.predictions_dict()) if output is not None else None
    )
    return prediction_log_pb2.ModelOutput(
        prediction_results_table=prediction_results_table
    )



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
from http import HTTPStatus
from typing import Any, Dict

import orjson
import tornado.web

from pyfuncserver.model.model import PyFuncModel
from merlin.pyfunc import PyFuncOutput


class PredictHandler(tornado.web.RequestHandler):
    def initialize(self, models):
        self.models = models  # pylint:disable=attribute-defined-outside-init
        self.publisher = self.application.publisher

    def get_model(self, full_name: str):
        if full_name not in self.models:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.NOT_FOUND,
                reason="Model with full name %s does not exist." % full_name
            )
        model = self.models[full_name]
        if not model.ready:
            model.load()
        return model

    def get_headers(self, request):
        return {k: v for (k, v) in request.headers.get_all()}

    def validate(self, request):
        try:
            body = orjson.loads(request.body)
        except json.decoder.JSONDecodeError as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.BAD_REQUEST,
                reason="Unrecognized request format: %s" % e
            )
        return body

    def post(self, full_name: str):
        model = self.get_model(full_name)

        request = self.validate(self.request)
        headers = self.get_headers(self.request)

        output = model.predict(request, headers=headers)
        response_json = output
        output_is_pyfunc_output = isinstance(response_json, PyFuncOutput)
        if output_is_pyfunc_output:
            response_json = output.http_response
         
        response_json = orjson.dumps(response_json)
        self.write(response_json)
        self.set_header("Content-Type", "application/json; charset=UTF-8")

        if self.publisher is not None and output_is_pyfunc_output and output.contains_prediction_log():
            tornado.ioloop.IOLoop.current().spawn_callback(self.publisher.publish, output)

    def write_error(self, status_code: int, **kwargs: Any) -> None:
        logging.error(self._reason)
        self.write({"status_code": status_code, "reason": self._reason})


class LivenessHandler(tornado.web.RequestHandler):  # pylint:disable=too-few-public-methods
    def get(self):
        self.write("Alive")


class HealthHandler(tornado.web.RequestHandler):
    def initialize(self, models: Dict[str, PyFuncModel]):
        self.models = models  # pylint:disable=attribute-defined-outside-init

    def get(self, full_name: str):
        if full_name not in self.models:
            raise tornado.web.HTTPError(
                status_code=404,
                reason="Model with full name %s does not exist." % full_name
            )

        model = self.models[full_name]
        self.write(json.dumps({
            "name": model.full_name,
            "ready": model.ready
        }))
